/*****************************************************************************
 * Copyright (c) 2006-2007, Cloudsmith Inc.
 * The code, documentation and other materials contained herein have been
 * licensed under the Eclipse Public License - v 1.0 by the copyright holder
 * listed above, as the Initial Contributor under such license. The text of
 * such license is available at www.eclipse.org.
 *****************************************************************************/
package org.eclipse.buckminster.ui.wizards;

import org.eclipse.buckminster.ui.Messages;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IProjectNature;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.dialogs.ContainerSelectionDialog;

public abstract class NewBMFileWizardPage extends WizardPage
{

	private Text containerText;

	private Text fileText;

	private String m_fileName;

	protected ISelection m_selection;

	private String m_extension;

	protected NewBMFileWizardPage(ISelection selection, String fileName, String extension)
	{
		super("wizardPage"); //$NON-NLS-1$
		m_selection = selection;
		m_fileName = fileName;
		m_extension = extension;
	}

	/**
	 * @see IDialogPage#createControl(Composite)
	 */
	public void createControl(Composite parent)
	{
		Composite container = new Composite(parent, SWT.NULL);
		GridLayout layout = new GridLayout();
		container.setLayout(layout);
		layout.numColumns = 3;
		layout.verticalSpacing = 9;
		Label label = new Label(container, SWT.NULL);
		label.setText(Messages.container_with_hotkey);

		containerText = new Text(container, SWT.BORDER | SWT.SINGLE);
		GridData gd = new GridData(GridData.FILL_HORIZONTAL);
		containerText.setLayoutData(gd);
		containerText.addModifyListener(new ModifyListener()
		{
			public void modifyText(ModifyEvent e)
			{
				dialogChanged();
			}
		});

		Button button = new Button(container, SWT.PUSH);
		button.setText(Messages.browse_with_dots);
		button.addSelectionListener(new SelectionAdapter()
		{
			@Override
			public void widgetSelected(SelectionEvent e)
			{
				handleBrowse();
			}
		});
		label = new Label(container, SWT.NULL);
		label.setText(Messages.file_name_with_hotkey);

		fileText = new Text(container, SWT.BORDER | SWT.SINGLE);
		gd = new GridData(GridData.FILL_HORIZONTAL);
		fileText.setLayoutData(gd);
		fileText.addModifyListener(new ModifyListener()
		{
			public void modifyText(ModifyEvent e)
			{
				dialogChanged();
			}
		});
		initialize();
		dialogChanged();
		setControl(container);
	}

	public String getContainerName()
	{
		return containerText.getText();
	}

	/**
	 * Returns the entered filename with correct extension.
	 * 
	 * @return
	 */
	public String getFileName()
	{
		String fileName = fileText.getText();
		if(!fileName.endsWith("." + m_extension)) //$NON-NLS-1$
			fileName += "." + m_extension; //$NON-NLS-1$
		return fileName;
	}

	/**
	 * Ensures that both text fields are set.
	 */
	private void dialogChanged()
	{
		IResource container = ResourcesPlugin.getWorkspace().getRoot().findMember(new Path(getContainerName()));
		String fileName = fileText.getText();

		if(getContainerName().length() == 0)
		{
			updateStatus(Messages.file_container_must_be_specified);
			return;
		}
		if(container == null || (container.getType() & (IResource.PROJECT | IResource.FOLDER)) == 0)
		{
			updateStatus(Messages.file_container_must_exist);
			return;
		}
		if(!container.isAccessible())
		{
			updateStatus(Messages.project_must_be_writable);
			return;
		}
		if(fileName.length() == 0)
		{
			updateStatus(Messages.file_name_must_be_specified);
			return;
		}
		if(fileName.replace('\\', '/').indexOf('/', 1) > 0)
		{
			updateStatus(Messages.file_name_must_be_valid);
			return;
		}
		int dotLoc = fileName.lastIndexOf('.');
		if(dotLoc != -1)
		{
			String ext = fileName.substring(dotLoc + 1);
			if(ext.equalsIgnoreCase(m_extension) == false)
			{
				updateStatus(NLS.bind(Messages.file_extension_must_be_0, m_extension));
				return;
			}
		}
		updateStatus(null);
	}

	/**
	 * Uses the standard container m_selection dialog to choose the new value for the container field.
	 */
	private void handleBrowse()
	{
		ContainerSelectionDialog dialog = new ContainerSelectionDialog(getShell(), ResourcesPlugin.getWorkspace()
				.getRoot(), false, Messages.select_new_file_container);
		if(dialog.open() == Window.OK)
		{
			Object[] result = dialog.getResult();
			if(result.length == 1)
			{
				containerText.setText(((Path)result[0]).toString());
			}
		}
	}

	/**
	 * Tests if the current workbench m_selection is a suitable container to use.
	 */
	private void initialize()
	{
		if(m_selection != null && m_selection.isEmpty() == false && m_selection instanceof IStructuredSelection)
		{
			IStructuredSelection ssel = (IStructuredSelection)m_selection;
			if(ssel.size() > 1)
				return;
			Object obj = ssel.getFirstElement();

			if(obj instanceof IProjectNature)
			{
				obj = ((IProjectNature)obj).getProject();
			}

			if(obj instanceof IResource)
			{
				IContainer container;
				if(obj instanceof IContainer)
					container = (IContainer)obj;
				else
					container = ((IResource)obj).getParent();
				containerText.setText(container.getFullPath().toString());
			}
		}
		fileText.setText(m_fileName);
	}

	private void updateStatus(String message)
	{
		setErrorMessage(message);
		setPageComplete(message == null);
	}

}
